/*********************************************************************
 * Desktop Station one R3 Sketch for Mr Push R3
 *
 * Copyright (C) 2017 Yaasan
 *
 */

// include the library code:
#include "Wire.h"
#include <EEPROM.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "DSCoreM_Type.h"
#include "DSCoreM_Common.h"
#include "DSCoreM_List.h"
#include "DSCoreM_DCC.h"
#include "DSCoreM_MM2.h"
#include "DSCoreM.h"
#include "Wire.h"
#include "Functions.h"
#include "NewSerial.h"
#include "TrackReporterS88_DS.h"


#define MAX_S88DECODER 1
#define RELPYERROR_300 "300 Command error"
#define RELPYERROR_301 "301 Syntax error"
#define RELPYERROR_302 "302 receive timeout"
#define RELPYERROR_303 "303 Unknown error"
#define RELPYERROR_NONE ""

#define LEDSTATE_STOP 0
#define LEDSTATE_RUN 1
#define LEDSTATE_ERR 2

#define PIN_RUNLED	2
#define PIN_EDC A1
#define PIN_CURRENT A2


#define THRESHOLD_OC 186  // 1024d * (1 / 11) * 10V / 5V
#define THRESHOLD_EDC 375  // 1024d * (1 / 11) * 10V / 5V
#define THRESHOLD_OV  999  // 1024d * (1 / 11) * 20V / 5V

#ifdef __LGT8FX8P__
#define ADC_SHIFT 2
#else
#define ADC_SHIFT 0
#endif



NewSerial SerialDS;


/* Decralation classes */
DSCoreLib DSCore;
TrackReporterS88_DS reporter(MAX_S88DECODER);


/* LED functions */
byte gLED_State = LEDSTATE_STOP;
byte gLED_Toggle = 0;
byte gLED_Counter = 0;

/* Parser */
String request;
String function;
word arguments[4];
word numOfArguments;
boolean result;

/* Button */
uint8_t gLastButton = 1;

/* Offset Buffer */
word gCurOffset = 512;

//Task Schedule
unsigned long gPreviousL5 = 0; //   50ms interval
unsigned long gPreviousL8 = 0; // 1000ms interval
unsigned long gPreviousL9 = 0; // 2000ms interval



void printString(const char *s, int x, int y);
void ReplyPowerPacket(byte inPower);
void TogglePower();
byte ChangeStatePower(byte inPower);
word tuningCurOffset(void);

/***********************************************************************
* 
* Boot setup
* 
* 
************************************************************************/

void setup()
{
	DSCore.Init();

	//Pin mode
	pinMode(PIN_CURRENT, INPUT);//Current
	pinMode(PIN_EDC, INPUT);//Edc
	pinMode(PIN_RUNLED, OUTPUT);
	
	//analogReadResolution(10);
	
	
	/* LED turn on */
	changeLED();
	
	/* Clear data of the S88 decoder */
	reporter.refresh();
	

	/* Current Offset */
	gCurOffset = tuningCurOffset();
	
	/* Debug message for PC */
	SerialDS.Init();
	SerialDS.println("--------------------------------------");
	SerialDS.println("Desktop Station one over DSair2 HW    ");
	SerialDS.println("--------------------------------------");
	SerialDS.println("100 Ready");
	
	//Reset task
	gPreviousL5 = millis();
	gPreviousL8 = millis();
	
	
	sei();    //ENABLE INTERRUPTION
	
}


boolean parse() 
{

	String request = SerialDS.getRequest();

	char aTemp[48];
	request.toCharArray(aTemp, 30);

	int lpar = request.indexOf('(');
	if (lpar == -1) 
	{
		return false;
	}

	function = String(request.substring(0, lpar));
	function.trim();

	int offset = lpar + 1;
	int comma = request.indexOf(',', offset);
	numOfArguments = 0;
	
	while (comma != -1) 
	{
		String tmp = request.substring(offset, comma);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
		offset = comma + 1;
		comma = request.indexOf(',', offset);
	}

	int rpar = request.indexOf(')', offset);
	while (rpar == -1) 
	{
		return false;
	}

	if (rpar > offset) 
	{
		String tmp = request.substring(offset, rpar);
		tmp.trim();
		arguments[numOfArguments++] = stringToWord(tmp);
	}

	return true;
}

boolean dispatch() 
{
	byte aValue;
	byte aValueHigh;
	boolean aResult;
	char aTempStr[16];
	word aLocProtocol;
	byte aDir = 0;
	long aSpeed = 0;
	
	if ((function == "setLocoDirection") || ((function == "sLD")))
	{
		
		if( arguments[1] > 0)
		{
			//FWD=1,REV=2を FWD=0,REV=1に変換
			aDir = arguments[1] - 1;
			
			//異常値はREV固定
			if(aDir > 1)
			{
				aDir = 1;
			}
		}
		else
		{
			//FWD固定
			aDir = 0;
		}
		
		return DSCore.SetLocoDirection(arguments[0], aDir);
	
	}
	else if ((function == "setLocoFunction") || ((function == "sLF")))
	{
		
		return DSCore.SetLocoFunction(arguments[0], arguments[1] + 1, arguments[2]);
	
	}
	else if ((function == "setTurnout") || ((function == "sTO")))
	{
		
		return DSCore.SetTurnout(arguments[0], (byte)arguments[1]);
	
	}
	else if ((function == "setPower") || ((function == "sPW")))
	{
		
		if( arguments[0] == 1)
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_RUN;
		}
		else
		{
			/* 状態変更 */
			gLED_State = LEDSTATE_STOP;
		}
		
		
		/* LED状態の変更 */
		changeLED();
		
		return DSCore.SetPower(arguments[0]);
	
	}
	else if ((function == "setLocoSpeed") || ((function == "sLS")))
	{
		
		return DSCore.SetLocoSpeedEx(arguments[0], arguments[1], 2);
	}
	else if ((function == "reset") || ((function == "sRS")))
	{
		SerialDS.println("100 Ready");
		return true;
	}
	/* reset */
	else if ((function == "setPing") || ((function == "sPI")))
	{
		/* Type(2bytes), Version(2bytes), UID(4bytes) */
		SerialDS.println("@DSG,00,03,00,02,00,00,00,00");
		return true;
	}
	else if ((function == "getLocoConfig") || ((function == "gCV")))
	{
		aValue = 0;
		
		aResult = DSCore.ReadConfig( arguments[1], &aValue, 0, 0);
		
		//Result
		if( aResult != true)
		{
			sprintf(aTempStr, "Read Err");
		}
		else
		{
			sprintf(aTempStr, "   =>%3d", aValue);
		}

		SerialDS.print("@CV,");
		SerialDS.printDEC(arguments[0]);
		SerialDS.print(",");
		SerialDS.printDEC(arguments[1]);
		SerialDS.print(",");
		
		if( aResult == true)
		{
			SerialDS.printDEC(aValue);
		}
		else
		{
			SerialDS.printDEC(0xFFFF);
		}
		
		SerialDS.println(",");
		
		return true;
	}
	else if (function == "getS88")
	{
		int aMaxS88Num = MAX_S88DECODER;
		
		if( arguments[0] > 0)
		{
			aMaxS88Num = arguments[0];
		}
		
		//Send a S88 sensor reply 
		SerialDS.print("@S88,");

		reporter.refresh(aMaxS88Num);

		word aFlags = 0;

		for( int j = 0; j < aMaxS88Num; j++)
		{
		  aFlags = (reporter.getByte((j << 1) + 1) << 8) + reporter.getByte(j << 1);
		  
		  SerialDS.printHEX((word)aFlags);
		  SerialDS.print(",");
		}
			
		SerialDS.println("");
		
		return true;
	} /* getS88 */
	else if ((function == "setLocoConfig") || ((function == "sCV")))
	{
		
		return DSCore.WriteConfig_Dir( arguments[1], arguments[2]);
	}
	else
	{
		return false;
	}
}

void loop()
{
	uint8_t aErrNo = 0;
	word aEdc;
	int aCur;
	
	while(1)
	{
		aErrNo = 0;
		
		/* Reset watchdog timer. */
		//wdt_reset(); 
		
		//パルスジェネレータのスキャン
		DSCore.Scan();	
		
		char *aReplyText = RELPYERROR_NONE;
		
		if( (millis() - gPreviousL8) >= 200)
		{
			
			//LED Control
			taskLED();
			
			/* Sensor check */
			aEdc = analogRead(PIN_EDC);
			aCur = analogRead(PIN_CURRENT) - gCurOffset;
			
			if( aCur < 0)
			{
				aCur = -aCur;
			}
			
			/* OC check */
			if( aCur > THRESHOLD_OC )
			{
				aErrNo = 3;
			}
			
			/* OV check */
			if( aEdc > THRESHOLD_OV )
			{
				aErrNo = 2;
			}
			/* LV check */
			if( aEdc < THRESHOLD_EDC )
			{
				aErrNo = 1;
				
			}		
			/* エラー処理 */
			if( aErrNo > 0)
			{
				if( DSCore.IsPower() == 1)
				{
					PowerOffByErr(aErrNo);
				}
				
				/*
				SerialDS.print("Edc=");
				SerialDS.printDEC(aEdc);
				SerialDS.print(", Cur=");
				SerialDS.printDEC(aCur);
				SerialDS.println(",");
				*/
			}
		
			//Reset task
			gPreviousL8 = millis();
			
		}
		
		
		if( SerialDS.received())
		{
			
	  		if (parse()) {
				if (dispatch()) {
					aReplyText = "200 Ok";
				} else {
					//SerialDS.println(function);
					aReplyText = RELPYERROR_300;
				}
	  			
			} else {
				aReplyText = RELPYERROR_301;
			}
			
			/* Reply to Desktop Station */
			SerialDS.println(aReplyText);
			
			SerialDS.clearReceive();
		}
		else
		{
			/* Nothing to do */
			
		}
		

		
	}
}

byte ChangeStatePower(byte inPower)
{
	byte aPower = inPower;
	
	if( aPower == 0)
	{
		aPower = 1;
		
		/* 状態変更 */
		gLED_State = LEDSTATE_RUN;
		
	}
	else
	{
		aPower = 0;
		
		/* 状態変更 */
		gLED_State = LEDSTATE_STOP;
	}
	
	return aPower;
}

void TogglePower()
{
	char aTempStr[16];
	uint8_t aPower = DSCore.IsPower();
	
	aPower = ChangeStatePower(aPower);
	
	DSCore.SetPower(aPower);
	ReplyPowerPacket(aPower);
	
	/* LED状態の変更 */
	changeLED();
}

void PowerOffByErr(uint8_t inErrNo)
{
	
	
	DSCore.SetPower(0);
	delay(1000);
	ReplyPowerPacket(0);
	
	SerialDS.print("@ERR,");
	SerialDS.printHEX(inErrNo);
	SerialDS.println(",");
	
	/* 状態変更 */
	gLED_State = LEDSTATE_ERR;
}

void ReplyPowerPacket(byte inPower)
{
	/* Power */
	SerialDS.print("@PWR,");
	SerialDS.printHEX(inPower);
	SerialDS.println(",");
}


void changeLED(void)
{
	
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		digitalWrite(PIN_RUNLED, 0);
		
		break;
		
	case LEDSTATE_RUN:
		digitalWrite(PIN_RUNLED, 1);
		break;
		
	case LEDSTATE_ERR:
		break;
	}
}

void taskLED(void)
{
	/* カウンタを回す(200msで1カウント) */
	gLED_Counter++;
	gLED_Counter = gLED_Counter & 0b1111;
	
	switch( gLED_State)
	{
	case LEDSTATE_STOP:
		
		if( gLED_Counter <= 3)
		{
			digitalWrite(PIN_RUNLED, 1);
		}
		else
		{
			digitalWrite(PIN_RUNLED, 0);
		}
		break;
		
	case LEDSTATE_RUN:
		break;
		
	case LEDSTATE_ERR:
		digitalWrite(PIN_RUNLED, gLED_Toggle);
		gLED_Toggle = (~gLED_Toggle) & 1;
		
		break;
	}
}

/* 電流のオフセット点検出 */
word tuningCurOffset(void)
{
	long aCurBuf = 0;
	
	for( int i = 0; i < 32; i++)
	{
		
		aCurBuf = aCurBuf + analogRead(A2);
		delay(5);
	}
	
	return aCurBuf >> 5;
	
	
}

